/*****************************************************************************/
//
// FILE        :  Image.cc
// AUTHOR      :  Jacco Vink 
// DESCRIPTION :  Implementation of Image class
//
//
// COPYRIGHT   :  Jacco Vink, AIP, 1999
//
/*****************************************************************************/
#include "Image.h"


/*****************************************************************************/
//
// FUNCTION    : Image()
// DESCRIPTION : constructor
// 
//
// SIDE EFFECTS:
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//
//  RETURNS    :
//
/*****************************************************************************/
Image::Image( void  )
{
  if( strcmp(classid,"Image")== 0){
    // warning("constructor(): Constructor illegaly called!");
    return;
  }
  strcpy(classid,"Image");

  reset_keywords();
  drdelt1 = 0.0;
  drdelt2 = 0.0;
  drpix1 = 0.0;
  drpix2 = 0.0;
  
  nx  = 0;
  ny  = 0;
  ncomplex = 0;
  pixel =0;
  cpixel = 0;

  ftransformed = false;
  fftplanexist = false;
  fftwdir =  FFTW_REAL_TO_COMPLEX;

  imgbin = 1;

}// end Image()


/*****************************************************************************/
//
// FUNCTION    : Image()
// DESCRIPTION : constructor
// 
//
// SIDE EFFECTS:
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//   xsize, nysize          I                     x/y size
//
//  RETURNS    :
//
/*****************************************************************************/
Image::Image( int xsize, int ysize, float val )
{
  if( strcmp(classid,"Image")== 0){
    // warning("constructor(): Constructor illegaly called!");
    return;
  }
  strcpy(classid,"Image");

  reset_keywords();
  drdelt1 = 0.0;
  drdelt2 = 0.0;
  drpix1 = 0.0;
  drpix2 = 0.0;
  
  nx  = 0;
  ny  = 0;
  ncomplex = 0;
  pixel = 0;
  cpixel = 0;

  ftransformed = false;
  fftplanexist = false;
  imgbin = 1;

  allocate( xsize, ysize, val );

}// end Image()

/*****************************************************************************/
//
// FUNCTION    : Image()
// DESCRIPTION : copy constructor
// 
//
// SIDE EFFECTS:
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//
//
//  RETURNS    :
//
/*****************************************************************************/
Image::Image( const Image &im )
{

  if( this == &im || strcmp(classid,"IMAGE")== 0){
    //    warning("Astroimage()/copy: Constructor illegaly called!");
    Image();
  }

  nx  = 0;
  ny  = 0;
  ncomplex = 0;
  pixel= 0;
  cpixel = 0;

  ftransformed = false;
  fftplanexist = false;
  fftwdir = FFTW_REAL_TO_COMPLEX;
  strcpy(classid,"IMAGE");

  if( im.nx > 0 && im.ny > 0 ){

    if( allocate( im.nx, im.ny ) == FAILURE ){
      //      warning("Astroimage(const Astroimage &im): Allocation error!");
    }
    else{
      
      copy_keywords( im ); // copy the keywords from object im
      imgbin = im.imgbin;
      
      long nxy = nx*ny;
      float *p1 = &pixel[0], *p2 = &im.pixel[0];
      for( long i = 0; i < nxy ; i++ ){
	*p1++ = *p2++;
      }// end for
      
    }//else	
  }
  //  else
  //    puts("error!");
    
}// end Image()


/*****************************************************************************/
//
// FUNCTION    : ~Image()
// DESCRIPTION : destructor
// 
//
// SIDE EFFECTS:
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//
//
//  RETURNS    :
//
/*****************************************************************************/
Image::~Image()
{

    deallocate();

    strcpy(classid,"ENDED");

}// end ~Image()

/*****************************************************************************/
//
// FUNCTION    :  operator=()
// DESCRIPTION :  copy operator. Only the image parameters are copied, 
//		not the event list.
// 
//
// SIDE EFFECTS:
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//
//
//  RETURNS    : reference to the object
//
/*****************************************************************************/
Image &Image::operator=(Image const &im )
{

  if( this == &im ){
    //  warning("operator=(): Cannot assign to itself!");
    return *this;
  }//if

  if( this!= &im && (im.nx != this->nx) || (im.ny != this->ny) ){
    if( allocate( im.nx, im.ny ) == FAILURE ){
      //warning("operator=(): Allocation error!");
      return *this;
    } // end if
  }// end if
  
  copy_keywords( im ); // copy the keywords from object im
  long nxy = nx*ny;
  float *p1 = pixel;
  float *p2 = im.pixel;
  for( long i = 0; i < nxy ; i++ ){
    *p1++ = *p2++;
  }// end for

  if( fftplanexist == true){
    rfftwnd_destroy_plan( fftplan);
    fftplanexist = false;
    fftwdir = FFTW_REAL_TO_COMPLEX;
  }

  return *this;
}// operator=()


/*****************************************************************************/
//
// FUNCTION    : allocate()
// DESCRIPTION : Allocates memory for the Image().
// 
//
// SIDE EFFECTS:
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//       n               I                 number of rows
//       m               I                           columns
//
//  RETURNS    :  SUCCESS or FAILURE
//
/*****************************************************************************/
Status Image::allocate( unsigned int n, unsigned int m, float initval)
{

  if( n <= 0 || m <= 0 || n > MAX_SIZE || m > MAX_SIZE ){
    // warning("allocate(): Error in the input parameters.");
    return FAILURE;
  }

  long nxy = n*m;

  if( nx != n || ny != m ){
    deallocate();
    pixel = new float [nxy]; // allocate
  }//if

  nx = n;
  ny = m;

  if( pixel == 0 ){ // memory allocated?
    nx = ny = 0;
    // warning("allocate(): Allocation error!");
    return FAILURE;
  }// end if

  for(long i = 0; i < nxy; i++ ){
    pixel[i] = initval;
  }//end for j


  return SUCCESS;

}// end allocate()
/*****************************************************************************/
//
// FUNCTION    : allocate_complex()
// DESCRIPTION : Allocates memory for the complex array and creates a
//               plan for the FFT (see manual of fftw).
//               The size of the array depends on nx, ny of the real image.
//               NOTE that fftw uses a different convention for the x and y
//               coordinates, y is the fastest changin coordinate.
//               So the naming convention for x and y are interchanged!
//
// SIDE EFFECTS:
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//
//
//  RETURNS    :  SUCCESS or FAILURE
//
/*****************************************************************************/
Status Image::allocate_complex( fftw_direction dir )
{

  if( nx <= 0 || ny <= 0 || nx > MAX_SIZE || ny > MAX_SIZE ){
    //    warning("allocate_complex(): Error in the input parameters.");
    return FAILURE;
  }

    // Note that nx is here the fastest running index
    // not ny (as in the FFTW manual)!
    //
  if( ncomplex != (unsigned long)(nx/2+1)*ny ){

    if( ncomplex != 0 || fftplanexist == true){
      fftplanexist = false;
      rfftwnd_destroy_plan( fftplan);
      delete [] cpixel;
    }//if

    // see note above
    ncomplex = (nx/2+1)*ny;
    cpixel = new fftw_complex [ncomplex];

    // create the plan
    fftplan = rfftw2d_create_plan(ny, nx, dir, FFTW_ESTIMATE);
    fftwdir = dir;
    fftplanexist = true;

  }
  else
    if( fftplanexist == false ){ //this can happen ncomplex=okay, fftwplan not
      // create the plan
      fftplan = rfftw2d_create_plan(ny, nx, dir, FFTW_ESTIMATE);
      fftwdir = dir;
      fftplanexist = true;
    }

  ftransformed = false;

  return SUCCESS;
}// end allocate_complex()


/*****************************************************************************/
//
// FUNCTION    : deallocate()
// DESCRIPTION : Deallocates the Image.
// 
//
// SIDE EFFECTS:
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//
//
//  RETURNS    :  SUCCESS
//
/*****************************************************************************/
Status Image::deallocate()
{

  if( nx > 0 && ny > 0 ){
    delete [] pixel;
    pixel = 0;
  } // end if
  nx = ny = 0;


  if( ncomplex > 0 ){
    delete [] cpixel;
    cpixel = 0;
    ftransformed = false;
  }
  ncomplex = 0;

  if(   fftplanexist == true){
    rfftwnd_destroy_plan( fftplan);
    fftplanexist = false;
  }

  return SUCCESS;
}// end deallocate()


/****************************************************************************/
//
// FUNCTION    : reset_keywords() 
// DESCRIPTION : 
// 
//
// SIDE EFFECTS:
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//
//
//  RETURNS    : SUCCESS
//
/****************************************************************************/
void Image::reset_keywords()
{

  //  warning("Entering reset_keywords()");

  //  strncpy(instrument,"UNKNOWN", FLEN_VALUE);
  instrument = string("UNKNOWN");
  telescope  = string("UNKNOWN");
  detnam     = string("UNKNOWN");
  datamode   = string("UNKNOWN");
  filter     = string("UNKNOWN");
  obs_mode   = string("UNKNOWN");
  obs_id     = string("UNKNOWN");
  exp_id     = string("UNKNOWN");
  object     = string("UNKNOWN");
  obsdate    = string("UNKNOWN");
  enddate    = string("UNKNOWN");
  observer   = string("UNKNOWN");

  ctype1     = string("UNKNOWN");
  ctype2     = string("UNKNOWN");
  radecsys   = string("FK5");
  pradecsys  = string("FK5");

  dec_nom   = 0.0;
  ra_nom    = 0.0;
  bscale    = 1.0;
  bzero     = 0.0;
  crval1    = 0.0;
  crval2    = 0.0;
  crpix1    = 0.0;
  crpix2    = 0.0;
  cdelt1    = 0.0;
  cdelt2    = 0.0;
  crota2    = 0.0;
  pc001001  = 1.0;
  pc002002  = 1.0;
  pc001002  = 0.0;
  pc002001  = 0.0;
  longpole  = 180.0;
  blank     = 0.0;
  equinox   = 2000.0;
  exposure  = 0.0;
  deadtime_corr = 1.0;
  ontime   = 0.0;
  start_time = 0.0;
  stop_time = 0.0;
  mjd_obs   = 0.0;
  chan_min = 0;
  chan_max = 2047;
  projectiontype = UNKNOWN;
  xy_type = DET_XY;


  ra_pnt    = 0.0;
  dec_pnt   = 0.0;
  pa_pnt    = 0.0;
  pequinox  = 2000.0;
  pexposure = 0.0;

  //  nhistory =0;

  return;
}// reset_keywords( )


/****************************************************************************/
//
// FUNCTION    : copy_keywords() 
// DESCRIPTION : 
// 
//
// SIDE EFFECTS:
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//
//
//  RETURNS    : SUCCESS
//
/****************************************************************************/
Status Image::copy_keywords( Image const &obs)
{

  if( this != &obs ){
    instrument = obs.instrument;
    telescope  = obs.telescope;
    detnam     = obs.detnam;
    datamode   = obs.datamode;
    filter     = obs.filter;
    obs_mode   = obs.obs_mode;
    obs_id     = obs.obs_id;
    exp_id     = obs.exp_id;
    object     = obs.object;
    obsdate    = obs.obsdate;
    enddate    = obs.enddate;
    ctype1     = obs.ctype1;
    ctype2     = obs.ctype2;
    radecsys   = obs.radecsys;
    pradecsys   = obs.pradecsys;
    

    dec_nom   = obs.dec_nom;
    ra_nom    = obs.ra_nom;
    crval1    = obs.crval1;
    crval2    = obs.crval2;
    crpix1    = obs.crpix1; 
    crpix2    = obs.crpix2;
    cdelt1    = obs.cdelt1;
    cdelt2    = obs.cdelt2;
    crota2    = obs.crota2;
    pc001001  = obs.pc001001;
    pc002002  = obs.pc002002;
    pc001002  = obs.pc001002;
    pc002001  = obs.pc002001;
    longpole  = obs.longpole;
    drpix1    = obs.drpix1;
    drpix2    = obs.drpix2;
    drdelt1   = obs.drdelt1;
    drdelt2   = obs.drdelt2;
    roll_nom  = obs.roll_nom;
    blank     = obs.blank;
    bscale    = obs.bscale;
    bzero     = obs.bzero;
    equinox   = obs.equinox;
    exposure  = obs.exposure;
    pequinox   = obs.pequinox;
    pexposure  = obs.pexposure;
    deadtime_corr = obs.deadtime_corr;
    ontime    = obs.ontime;
    mjd_obs   = obs.mjd_obs;
    chan_min  = obs.chan_min;
    chan_max  = obs.chan_max;
    ra_pnt    = obs.ra_pnt;
    dec_pnt   = obs.dec_pnt;
    pa_pnt    = obs.pa_pnt;
    projectiontype = obs.projectiontype;

  }
  else{
    //    warning("copy_keywords(): Cannot copy to the same object!");
    return FAILURE;
  }//else


  return SUCCESS;
}// copy_keywords( )





/*****************************************************************************/
//
// FUNCTION    : operator+=()
// DESCRIPTION : Adds an other image to the image.
// 
//
// SIDE EFFECTS:
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//      im                I                      image
//
//  RETURNS    : SUCCESS or FAILURE
//
/*****************************************************************************/
Status Image::operator+=( const Image &im )
{
    if( nx != im.nx || ny != im.ny || nx < 1  || ny < 1 ){
      //        warning("operator+=(im): The number of pixels do not match!");
        return FAILURE;
    }// end if

    if( cdelt1 != im.cdelt1 ||  cdelt2 != im.cdelt2 || crval1 != im.crval1 ||
	 crval2 != im.crval2 || crpix1 != im.crpix1 || crpix2 != im.crpix2 ||
	equinox!=im.equinox ){
      Sas::warning("diffCoord","World coordinates do not match!");
    }

    float *ptr1 = pixel;
    float *ptr2 = im.pixel;
    long nxy = nx*ny;
    for( long i = 0; i < nxy ; i++ ){

      if( ( *ptr1 == blank && blank != 0.0) || 
	  ( *ptr2 == im.blank && im.blank!= 0.0)){
	*ptr1 = blank;
      }
      else{
	*ptr1 += *ptr2;
      }//if
      ptr1++;
      ptr2++;

    }// end for

    if( exposure > 0.0 && im.exposure > 0.0){
      exposure += im.exposure;
    }

    //    reset_detector_keywords();
    return SUCCESS;

}// end operator+=() 


/*****************************************************************************/
//
// FUNCTION    : operator-=()
// DESCRIPTION : Subtracts an other image from the image.
// 
//
// SIDE EFFECTS:
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//      im                I                      image
//
//  RETURNS    : SUCCESS or FAILURE
//
/*****************************************************************************/
Status Image::operator-=( const Image &im )
{
  if( nx != im.nx || ny != im.ny || nx < 1  || ny < 1 ){
    //    warning("operator-=(im): The number of pixels do not match!");
    return FAILURE;
  }// end if

  if( cdelt1 != im.cdelt1 ||  cdelt2 != im.cdelt2 || crval1 != im.crval1 ||
      crval2 != im.crval2 || crpix1 != im.crpix1 || crpix2 != im.crpix2  ||
      equinox !=im.equinox ){
    Sas::warning("diffCoord","World coordinates do not match!");
  }

  float *ptr1 = pixel;
  float *ptr2 = im.pixel;
  long nxy = nx*ny;
  for( long i = 0; i < nxy ; i++ ){
      
      if( ( *ptr1 == blank && blank != 0.0) || 
	  ( *ptr2 == im.blank && im.blank != 0.0)){
	*ptr1 = blank;
      }
      else{
	*ptr1 -= *ptr2;
      }//if
      ptr1++;
      ptr2++;

  }// end for
  
  if( exposure > 0.0 && im.exposure > 0.0){
    exposure -= im.exposure;
  }

  //  reset_detector_keywords();
  return SUCCESS;

}// end operator-=() 


/*****************************************************************************/
//
// FUNCTION    : operator*=()
// DESCRIPTION : Multiplies an other image with the image.
// 
//
// SIDE EFFECTS:
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//      im                I                      image
//
//  RETURNS    : SUCCESS or FAILURE
//
/*****************************************************************************/
Status Image::operator*=( const Image &im )
{
  if( nx != im.nx || ny != im.ny || nx < 1  || ny < 1 ){
    //    warning("operator*=(im): The number of pixels do not match!");
    return FAILURE;
  }// end if


  if( cdelt1 != im.cdelt1 ||  cdelt2 != im.cdelt2 || crval1 != im.crval1 ||
      crval2 != im.crval2 || crpix1 != im.crpix1 || crpix2 != im.crpix2 ||
      equinox!=im.equinox ){
    Sas::warning("diffCoord","World coordinates do not match!");
  }
  
  float *ptr1 = pixel;
  float *ptr2 = im.pixel;
  long nxy = nx*ny;
  for( long i = 0; i < nxy ; i++ ){
      
    if( ( *ptr1 == blank && blank != 0.0) || 
	( *ptr2 == im.blank && im.blank!= 0.0)){
      *ptr1 = blank;
    }
    else{
      *ptr1 *= *ptr2;
    }//if
    ptr1++;
    ptr2++;
    
  }// end for

  // reset_detector_keywords();

  return SUCCESS;

}// end operator*=() 

/*****************************************************************************/
//
// FUNCTION    : operator/=()
// DESCRIPTION : Devides the image by an another image. Checks for division by
//		 zero.
// 
//
// SIDE EFFECTS:
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//      im                I                      image
//
//  RETURNS    : SUCCESS or FAILURE
//
/*****************************************************************************/
Status Image::operator/=( const Image &im )
{
  if( nx != im.nx || ny != im.ny || nx < 1  || ny < 1 ){
    //    warning("operator/=(im): The number of pixels do not match!");
    return FAILURE;
  }// end if
    
   
  if( cdelt1 != im.cdelt1 ||  cdelt2 != im.cdelt2 || crval1 != im.crval1 ||
      crval2 != im.crval2 || crpix1 != im.crpix1 || crpix2 != im.crpix2 ||
      equinox!=im.equinox ){
    Sas::warning("diffCoord","World coordinates do not match!");
  }

  bool zerodivision = false;
  float *ptr1 = pixel;
  float *ptr2 = im.pixel;
  
  long nxy = nx*ny;
  for( long i = 0; i < nxy ; i++ ){   

    if( ABS( *ptr2 ) > MIN_DIV && 
	( *ptr2  != im.blank || im.blank==0.0) && 
	( *ptr1 != blank || blank == 0.0 ) ){
      *ptr1 /=  *ptr2;
    }
    else{
      *ptr1 = blank;
      zerodivision = true;
    }// end if
    ptr1++;
    ptr2++;

  }// end for

  //  reset_detector_keywords();

  if( zerodivision == false ){
    return SUCCESS;
  }
  else{
    //    warning("operator/=(im): Divison by zero encountered!");
    return FAILURE;
  }// end if

}// end operator/=()

/*****************************************************************************/
//
// FUNCTION    : operator*=(double )
// DESCRIPTION : Multiplies all pixels by a constant.
// 
//
// SIDE EFFECTS:
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//       a              I                         float value
//
//  RETURNS    :  SUCCESS or FAILURE
//
/*****************************************************************************/
Status Image::operator*=( double a )
{
  if( nx < 1  || ny < 1 ){
    //    warning("operator*=(): This image has no pixels!");
    return FAILURE;
  }// end if
  
  long nxy = nx*ny;
  for( long i = 0; i < nxy ; i++ ){
    if( pixel[i] != blank || blank==0.0){
      pixel[i]*= a;
    }
  }// end for


  //exposure *= a;  


  return SUCCESS;

}// end operator*=() 

/*****************************************************************************/
//
// FUNCTION    : operator/=()
// DESCRIPTION : Divides by a value.
// 
//
// SIDE EFFECTS:
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//       a              I                         float value
//
//  RETURNS    :  SUCCESS or FAILURE
//
/*****************************************************************************/
Status Image::operator/=( double a )
{
  if( nx < 1  || ny < 1  || a == 0.0 ){
    //  warning("operator/=(): This image has no pixels or division by zero!");
    return FAILURE;
  }// end if

  long nxy = nx*ny;
  for( long i = 0; i < nxy ; i++ ){
    if( pixel[i] != blank || blank==0.0 ){
      pixel[i] /= a;
    }
  }// end for

  exposure /= a;

  return SUCCESS;

}// end operator/=() 

/*****************************************************************************/
//
// FUNCTION    : operator+=()
// DESCRIPTION : Adds a constant to all pixels.
// 
//
// SIDE EFFECTS:
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//       a              I                         float value
//
//  RETURNS    :  SUCCESS or FAILURE
//
/*****************************************************************************/
Status Image::operator+=( double a )
{
  if( nx < 1  || ny < 1 ){
    //    warning("operator+=(): This image has no pixels!");
    return FAILURE;
  }// end if
  
  long nxy = nx*ny;
  for( int i = 0; i < nxy ; i++ ){
        if( pixel[i] != blank || blank == 0.0 ){
	  pixel[i]+= (float)a;
        }
  }// end for

  return SUCCESS;

}// end operator+=() 

/*****************************************************************************/
//
// FUNCTION    : total()
// DESCRIPTION : Calculates the integrated value. The sommation is 
//               internally done using a long double.
// 
//
// SIDE EFFECTS:
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//
//
//  RETURNS    :
//
/*****************************************************************************/
double Image::total( void ) const
{
    if( nx < 1 || ny < 1 ){
      //warning("total(): The sum is not defined!");
      return 0.0;
    }// end if

    double sum=0.0;
    long nxy = nx*ny;
    for( int i =0; i < nxy ; i++)
        if( pixel[i] != blank || blank == 0.0 ){
            sum += pixel[i];
        }// end if

    return sum;
}// total()

/*****************************************************************************/
//
// FUNCTION    : total()
// DESCRIPTION : Calculates the integrated value. The sommation is 
//               internally done using a long double.
// 
//
// SIDE EFFECTS:
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//
//
//  RETURNS    :
//
/*****************************************************************************/
double Image::total( unsigned int i1, unsigned int j1, 
		     unsigned int i2, unsigned int j2 ) const
{
    if( nx < 1 || ny < 1 || i2 < i1 || j2 < j1 || j1<= 0 || j2 <= 0 ||
	i2 > nx || j2 > ny){
      //        warning("total(): The sum is not defined!");
        return 0.0;
    }// end if

    i1 -= 1;//fits convention
    i2 -= 1;
    j1 -= 1;
    j2 -= 1;

    unsigned int i , j;
    double sum=0.0, pix;

    for( j  = j1; j <= j2 ; j++){
      for( i  = i1; i <= i2 ; i++){
	pix = pixel[i+j*nx];
        if( pix != blank || blank == 0.0){
            sum += pix;
        }// end if
      }
    }

    return sum;

}// total()

/*****************************************************************************/
//
// FUNCTION    : minimum()
// DESCRIPTION : Return the minimum pixel value.
// 
//
// SIDE EFFECTS:
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//    
//
//  RETURNS    :  minimum of pixel values
//
/*****************************************************************************/
float Image::minimum( void ) const
{
  if( nx < 1 || ny < 1 ){
    //    warning("minimum(): Pixel minimum is not defined!");
    return 0.0;
  }// end if

  long nxy = nx*ny;
  float min = pixel[ny/2];
  for( long i = 0; i < nxy ; i++){
    if( pixel[i] != blank || blank == 0.0){
      min =  MIN( min , pixel[i]);
    }// end i
  }// end for( int i 
  
  return min;
    
}// end minimum()


/*****************************************************************************/
//
// FUNCTION    : maximum()
// DESCRIPTION : Return the maximum pixel value.
// 
//
// SIDE EFFECTS: 
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//    
//
//  RETURNS    :  maximum of pixel values
//
/*****************************************************************************/
float Image::maximum( void ) const
{
    if( nx < 1 || ny < 1 ){
      //        warning("maximum(): Pixel minimum is not defined!");
        return 0.0;
    }// end if

    long nxy = nx*ny;
    float max = pixel[nxy/2];
    for( long i = 0; i < nxy ; i++){
      if( pixel[i] != blank || blank == 0.0){
           max =  MAX( max , pixel[i]);
      }// end i
    }// for

    return max;

}// maximum()

/*****************************************************************************/
//
// FUNCTION    : mean()
// DESCRIPTION : Return the mean pixel value.
// 
//
// SIDE EFFECTS: 
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//    
//
//  RETURNS    :  mean of non-blank pixel values
//
/*****************************************************************************/
float Image::mean( void ) const
{
    if( nx < 1 || ny < 1 ){
      //        warning("mean(): Mean value is not defined!");
        return 0.0;
    }// end if


    float sum=0.0;
    long nxy = nx*ny;
    float npixels = 0.0;
 
    for( long i = 0; i < nxy ; i++){
      if( pixel[i] != blank || blank == 0.0){
           sum =  sum + pixel[i];
           npixels = npixels + 1.0;
      }// end if
    }// for
    
    return sum/npixels;



}// mean()

/*****************************************************************************/
//
// FUNCTION    : clip()
// DESCRIPTION : Sets all pixel values of pixels with values outside a certain
//               range to BLANK.
// 
//
// SIDE EFFECTS:
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//   minval, maxval     I                the boundaries of the valid interval
//
//  RETURNS    : SUCCESS or FAILURE
//
/*****************************************************************************/
Status Image::clip(float minval, float maxval)
{
  if(minval >= maxval){
    //    warning("clip(): minval >= maxval!");
    return FAILURE;
  }// end clip()

  unsigned int nxy  = nx*ny;
  for(unsigned long i=0; i < nxy; i++ ){
    if( pixel[i] < minval || ( pixel[i] > maxval)  ){
      pixel[i] = 0.0;
    }//end if
  }// end for
  
  return SUCCESS;
}// end clip()


/*****************************************************************************/
//
// FUNCTION    : cage()
// DESCRIPTION : Sets all pixel values of pixels with values outside a certain
//               range to the margin values of the range.
// 
//
// SIDE EFFECTS:
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//   minval, maxval     I                the boundaries of the valid interval
//
//  RETURNS    : SUCCESS or FAILURE
//
/*****************************************************************************/
Status Image::cage(float minval, float maxval)
{
  if(minval >= maxval){
    //    warning("cage(): minval >= maxval!");
    return FAILURE;
  }// end cage()

  unsigned int nxy  = nx*ny;
  for(unsigned long i=0; i < nxy; i++ ){
    if( pixel[i] < minval ){
      pixel[i] = minval;
    }//end if
    if( pixel[i] > maxval ){
      pixel[i] = maxval;
    }//end if

  }// end for
  
  return SUCCESS;
}// end cage()

/*****************************************************************************/
//
// FUNCTION    : get()
// DESCRIPTION : Return the pixel value of pixel (i,j)
//               Note that the inputs follows the FITS convention, 
//               so the first pixel 1,1
//
// SIDE EFFECTS:
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//   i, j               I                 pixel position.
//
//  RETURNS    : pixel value
//
/*****************************************************************************/
float Image::get( unsigned int i, unsigned int j ) const
{
    if( i > nx || j > ny || i < 1  || j < 1 ){
      //        warning("get(): The requested coordinates fall outside the image!");
        return 0.0;
    }//end if

    return pixel[i-1 + (j-1)*nx];
}//end get()

/*****************************************************************************/
//
// FUNCTION    : getExposure()
// DESCRIPTION : Returns the exposure time.
// 
//
// SIDE EFFECTS:
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//
//
//  RETURNS    :  exposure time
//
/*****************************************************************************/
float Image::getExposure( void ) const
{

    return exposure;
}//end getExposure()

/*****************************************************************************/
//
// FUNCTION    : setpixel()
// DESCRIPTION : Sets a pixel to a new value
// 
//
// SIDE EFFECTS:
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//   i, j               I                 pixel position.
// val                  I                 new pixel value
//  RETURNS    : SUCCESS or FAILURE
//
/*****************************************************************************/
Status Image::setpixel( unsigned int i, unsigned int j, float val )
{
  if( i == 0 || i > nx || j == 0 || j> ny ){
    //    warning("set(): Pixel does not exist!");
    return FAILURE;
  }

  pixel[i-1 + (j-1)*nx] = val; // NOTE that we correct for 
                               // the FORTRAN-FITS convention

  return SUCCESS;
}// setpixel()

/*****************************************************************************/
//
// FUNCTION    : getnxny()
// DESCRIPTION : Return the size of the image
// 
//
// SIDE EFFECTS:
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//   m, n               O                nx and ny
//
//  RETURNS    : SUCCESS
//
/*****************************************************************************/
void Image::getnxny( unsigned int *m, unsigned int *n ) const
{

  *m= nx;
  *n = ny;
  return;
}//end getnxny()

/*****************************************************************************/
//
// FUNCTION    : windowfilter()
// DESCRIPTION : 
// 
//
// SIDE EFFECTS:
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//   window             I               subimage containing smearing function
//   n                  I
//   i1,j1,i2,j2        I                defines window, negatove numbers
//                                       indicate the whole image
//
//  RETURNS    : SUCCESS or FAILURE
//
/*****************************************************************************/
Status Image::windowfilter(const Image &im, float *window, 
			   int n, bool renormalize,
			   int i1, int j1, int i2, int j2)
{

  if( i1 == -1 && j1 == -1 && i2 == -1 && j2 == -1 ){
    i1 = j1 = 0;
    i2 = im.nx-1;
    j2 = im.ny-1;
  }//
  else{
    i1 -= 1; // FITS 2 C convention
    j1 -= 1;
    i2 -= 1;
    j2 -= 1;
  }//if

  if( n <= 0 || i2 <= i1 || j2 <= j1 || i1 < 0 || j1 < 0 || 
      i2 >= (int)im.nx || j2 >= (int)im.ny){
    //    warning("windowfilter(): Invalid parameters!");
    return FAILURE;
  }
  //  setblank( 0.0 );


  float norm, norm0;

  *this = im;

  norm0 = 0.0;

  register int k, l;
  if( renormalize == true){
    for( l= 0 ; l < n ; l++){
      for( k= 0 ; k < n ; k++){
	norm0 += window[ l*n + k];
	if( k != 0 && l !=0 )
	  norm0 += 3.0 * window[ l*n + k];
	if( (k==0 && l!=0) || (k!=0 && l==0))
	  norm0 += window[ l*n + k];
	
      }//for
    }//for
    if( norm0 == 0.0 ){
      return FAILURE;
    }
  }

  float winval, pixval, val, corr=1.0;
  int ipk, imk, jpl, jml,nl;
  int i, j;
  for(j = j1; j <= j2 ; j++){
    for(i = i1; i <= i2 ; i++ ){

      corr = 1.0;
      if( renormalize== true && 
	  (i-n <= i1 || j-n <= j1 || i+n >= i2 || j+n >= j2) ){

	norm = 0.0;
	for( l= 0 ; l < n ; l++){
	  nl = n*l;
	  for( k= 0 ; k < n ; k++){

	    winval= window[k + nl];
	    // first determine normalization:
	    if( i+k <= i2 ){
	      if( j+l <= j2)         // first quadrant
		norm += winval;
	      if( l != 0 && j-l >= j1 ) // fourth quadrant
		norm += winval;
	    }
	    if( k != 0 && i-k >= i1){
	      if( j+l <= j2)             // second quadrant
		norm += winval;

	      if( l != 0 && j-l >= j1 )  // third quadrant
		norm += winval;
	    }// if 

	  }//for
	}//for
	if( norm > 0.0 )
	  corr = norm0/norm;
	else
	  corr = 1.0;
      }//if

      pixval = 0.0;

      // Now make the convolution
      for( l= 0 ; l < n ; l++){
	nl = n*l;
	for( k= 0 ; k < n ; k++){

	  ipk = i+k; // for speed reasons: calculate only once
	  imk = i-k;
	  jpl = j+l;
	  jml = j-l;
	  val = window[ k + nl]*corr;
	  if( ipk <= i2 ){
	    if( jpl <= j2)         // first quadrant
	      pixval += im.pixel[ ipk + jpl * nx] * val;
	    if( jml >= j1 && l != 0 ) // fourth quadrant
	      pixval += im.pixel[ ipk + jml * nx] * val;	    
	  }

	  if( imk >= i1 && k != 0 ){
	    if( jpl <= j2)  // second quadrant
	      pixval += im.pixel[ imk + jpl * nx] * val;
	    if( jml >= j1 && l != 0) // third quadrant
	      pixval += im.pixel[ imk + jml * nx] * val;
	  }// if 
	}//for
      }//for

      pixel[i+j*nx] = pixval;

    }// for i
  }//for j


  return SUCCESS;

}// windowfilter()

/*****************************************************************************/
//
// FUNCTION    : gaussianfilter()
// DESCRIPTION : 
// 
//
// SIDE EFFECTS:
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
// sigma               I                         width of the gaussian
//
//  RETURNS    : SUCCESS or FAILURE
//
/*****************************************************************************/
Status Image::gaussianfilter(const Image &im,
			     float sigma, int i1, int j1, int i2, int j2)
{

  if( sigma > 32.0 || im.nx < 2 || im.ny < 2 ){
    //  warning("gaussianfilter(): The image is too small or sigma too big!");
    return FAILURE;
  }
  //  setblank( 0.0 );

  int i, j, n;
  float *window, norm, sum;

  //  n = (int)(4.0 * sigma + 1.0);
  n = (int)(3.5 * sigma + 1.51);
  window = new float [n*n];
  if( window == 0 ){
    //    warning("gaussianfilter(): Error allocating extra memory!");
    return FAILURE;
  }// window

  // Make a 2D Gaussian, but only half of it!
  sum=0.0;
  norm= 1.0/(2.0 * M_PI *sigma*sigma );
  for(j = 0 ; j < n ; j++){
    for( i = 0; i < n ; i++){
      window[ j*n + i] = norm * exp( -0.5*( i*i + j*j)/(sigma*sigma) );
      sum += window[ j*n + i];
      if( i != 0 && j!=0 ){
	sum += 3.0 * window[ j*n + i];
      }
      if( (i==0 && j!=0) || (i!=0 && j==0)){
	sum += window[ j*n + i];
      }
    }// for
  }//for 

  if( windowfilter( im, window, n, true, i1, j1, i2, j2) == FAILURE ){
    delete [] window;
    //    warning("gaussianfilter(): Error encountered!");
    return FAILURE;
  }//if

  delete [] window;

  return SUCCESS;

}// gaussianfilter

//
// based on select
// from Numerical Recipes in C, Press et al.
#define SWAP(a,b) temp=(a);(a)=(b);(b)=temp;
inline float select(unsigned long k, unsigned long n, float arr[])
{
	unsigned long i,ir,j,l,mid;
	float a,temp;

	l=1;
	ir=n;
	for (;;) {
		if (ir <= l+1) {
			if (ir == l+1 && arr[ir] < arr[l]) {
				SWAP(arr[l],arr[ir])
			}
			return arr[k];
		} else {
			mid=(l+ir) >> 1;
			SWAP(arr[mid],arr[l+1])
			if (arr[l] > arr[ir]) {
				SWAP(arr[l],arr[ir])
			}
			if (arr[l+1] > arr[ir]) {
				SWAP(arr[l+1],arr[ir])
			}
			if (arr[l] > arr[l+1]) {
				SWAP(arr[l],arr[l+1])
			}
			i=l+1;
			j=ir;
			a=arr[l+1];
			for (;;) {
				do i++; while (arr[i] < a);
				do j--; while (arr[j] > a);
				if (j < i) break;
				SWAP(arr[i],arr[j])
			}
			arr[l+1]=arr[j];
			arr[j]=a;
			if (j >= k) ir=j-1;
			if (j <= k) l=i;
		}
	}
}
#undef SWAP
/* (C) Copr. 1986-92 Numerical Recipes Software *;#1. */

/*****************************************************************************/
//
// FUNCTION    : medianfilter()
// DESCRIPTION : Takes the mdian value of n x n pixels. 
//               At the borders the pixels are wrapped around.
// 
//
// SIDE EFFECTS:
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//
//
//  RETURNS    : SUCCESS or FAILURE
//
/*****************************************************************************/
Status Image::medianfilter(unsigned int n)
{
  if( n % 2 == 0 || n <= 2 || n > MAX_MEDIAN || nx < 2*n || ny < 2*n ){
    //warning("medianfilter(): Need an odd input parameter, < 7!");
    return FAILURE;
  }
  //  setblank( 0.0 );

  float *window;
  Image newim;

  newim= *this;

  window = new float [n*n];
  if( window == 0 ){
    //    warning("medianfilter(): Error allocating extra memory!");
    return FAILURE;
  }// window



  int i, m,  lnx, jnx, nmedian = (n+1)/2 - 1;
  register int k, l;

  for(int j = 0; j < (int)ny ; j++){
    jnx =j * nx;
    for(i = 0; i < (int)nx ; i++ ){

      m = 0;
      for( l = j - nmedian ; l <= j + nmedian ; l++){
	lnx = l*nx;
	for( k = i - nmedian ; k <= i + nmedian; k++ ){

	  if( k >= 0 && k < (int)nx && l >= 0 && l < (int)ny)
	    window[m++] = pixel[k + lnx];

	}// for k
      }//for l
      newim.pixel[i + jnx ] = select((m+1)/2, m, window-1);
    }// for i
  }//for j


  delete [] window;
  *this = newim;

  return SUCCESS;

}// end medianfilter()


/*****************************************************************************/
//
// FUNCTION    : rebin()
// DESCRIPTION : Rebins the image with a value n>0 and nx,ny should be 
//               an integer value of nx, ny.
//               This will be replacedby a rebinned version of the input image.
//
// SIDE EFFECTS:
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//      im              I                 inout image to be rebinned
//      n               I                 rebinning factor
//
//  RETURNS    :  SUCCESS or FAILURE
//
/*****************************************************************************/
Status Image::rebin( const Image &im, unsigned int n )
{
  if( n < 1  || &im == this ){
    //    Sas::warning("rebinInv","Invalid input image for rebinning.");
    return FAILURE;
  }// end if

  if( im.nx % n != 0 || im.ny % n != 0 || im.nx*im.ny <= 0.0){
    //Sas::warning("wrongBinfac","image size cannot be divided by rebin factor");
    return FAILURE;
  }// end if


  if( n==1 ){
    //    Sas::message(Msg::App, Msg::Verbose,"No use to rebin with factor 1");
    return SUCCESS;
  }// end if

  if( allocate(im.nx/n, im.ny/n, 0.0) == FAILURE){
    //    Sas::warning("rebinAll","Error allocating new image for rebinning");
    return FAILURE;
  }
  copy_keywords( im );
  blank = im.blank;

  float newval;
  bool undef = false;
  unsigned int oldi, oldj;

  unsigned int i, j;
  register unsigned int  k, l;
  for( j = 0; j < ny ; j++){
      for( i = 0 ; i < nx ; i++){

	newval = 0.0;
	for( k = 0; k < n && !undef; k++ ){  // add neighbouring pixels
	  for( l = 0; l < n && !undef; l++ ){
	    oldi = (i*n + k);
	    oldj = (j*n + l);
	    if((im.pixel[ oldi + oldj * im.nx]== im.blank && im.blank!=0.0) || 
		oldi >= im.nx || oldj >= im.ny ){
	      undef = true; // undefined pixel value encountered   
	    }
	    else{
	      newval+= im.pixel[oldi + oldj*im.nx];
	    }// end if
	  }//end for(l
	}//end for(k
	if( i < nx && j < ny ){
	  if( undef == true ){
	    pixel[i + j*nx] = blank;     // assign value to the pixel
	    undef = false;
	  }
	  else{
	    pixel[i + j*nx] = newval;
	  }//end if
	}// end if

      }//end for j
    }//end for i


    // correct the scale keywords
    cdelt1 *= n;
    cdelt2 *= n;
    crpix1  = (crpix1-0.5)/n + 0.5; // FORTRAN convention first pixel=(1,1)
    crpix2  = (crpix2-0.5)/n + 0.5;
    imgbin *= n;

    return SUCCESS;
}// rebin()

/*****************************************************************************/
//
// FUNCTION    : getradec()
// DESCRIPTION : Return the AR and DEC of pixel x,y
// 
//
// SIDE EFFECTS: Note the first pixel is 1, 1 (0,0 in C++)
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//   x, y,                I
//   alpha, delta         O                 position in degrees.
//
//  RETURNS    : SUCCESS or FAILURE
//
/*****************************************************************************/
Status Image::getradec(double x, double y, double *alpha, double *delta) const
{
  if( cdelt1 == 0.0 || cdelt2 == 0.0){
    //    warning("getradec(): Physical coordinates not specified!");
    return FAILURE;
  }

  double phi, theta, rt, x0, y0;
  double ap = crval1*DEG2RAD, deltp = crval2 * DEG2RAD;
  double phip = longpole*DEG2RAD; 
  // pole of coordinate system

  *alpha = 0.0;
  *delta = 0.0;
  // first calculate phi, theta (i,j)
  //
  // (note that the pixels start by convention on 1 instead 0)

  // Inverse rotation:
  x0 = pc001001 * (x - crpix1) + pc001002 * (y - crpix2); 
  y0 = pc002001 * (x - crpix1) + pc002002 * (y - crpix2); 
  x0 *= cdelt1;
  y0 *= cdelt2;
  rt  = sqrt( x0*x0 + y0*y0 );               // Rtheta in GC
  phi = atan2( x0, -y0);                     // argument of x and y

  if( rt != 0.0 ){
    if( ctype1 == string("RA---TAN") && ctype2 == string("DEC--TAN") ){
      theta = atan( RAD2DEG/rt );
    }
    else{
      if( ctype1 == string("RA---SIN") && ctype2 == string("DEC--SIN") == 0){
	theta = acos( rt*DEG2RAD );
      }else{
     //warning("getradec(): No ctype keywords specified or not implemented.");
	return FAILURE;
      }//else
    }//else

    // calculate alpha and delta
    ch_coord(phi, theta, phip, ap, deltp, alpha, delta);
    *alpha *= RAD2DEG;
    *delta *= RAD2DEG;
  }
  else{
    //    warning("getradec(): theta not specified!");
    *alpha = crval1;
    *delta = crval2; 
    return FAILURE;
  }

  return SUCCESS;
}// getradec()

/*****************************************************************************/
//
// FUNCTION    : getxy()
// DESCRIPTION : Return the X,Y for a certain RA DEC
// 
//
// SIDE EFFECTS: Note that it is not checked if x and y are within the limits.
//               So the validity of the result is not checked.
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//   alpha, delta         I                 position in degrees.
//   x, y,                O
//
//  RETURNS    : SUCCESS or FAILURE
//
/*****************************************************************************/
Status Image::getxy( double alpha, double delta, 
		     double *x, double *y) const
{
  if( cdelt1 == 0.0 || cdelt2 == 0.0 ){
    //    warning("getxy(): Physical coordinates not specified!");
    return FAILURE;
  }
  
  double phi, theta, rt, x0, y0;
  double ap = crval1*DEG2RAD, deltp = crval2 * DEG2RAD;
  double phip = longpole*DEG2RAD; 
  double ipc001001, ipc001002, ipc002001, ipc002002, det;

  *x = 0.0;
  *y = 0.0;
  rt = 0.0;

  // make the inverted pc matrix:
  det = pc001001*pc002002 - pc002001*pc001002; // determinant of matrix
  if(det == 0.0 ){
    //    warning("getxy(): Error in the PC matrix!");
    return FAILURE;
  } //if
  ipc001001 =  pc002002/det; // inverted PC matrix
  ipc001002 = -pc001002/det;
  ipc002001 = -pc002001/det;
  ipc002002 =  pc001001/det;


  ch_coord( alpha*DEG2RAD, delta*DEG2RAD, ap, phip, deltp, &phi , &theta );

  if( theta != 0.0 ){
    if( ctype1 == string("RA---TAN") && ctype2 == string("DEC--TAN") ){
      rt = RAD2DEG/tan( theta );
    }
    else{
      if( ctype1 == string("RA---SIN") && ctype2 == string("DEC--SIN") ){
	rt = RAD2DEG * cos(theta);
      }
      else{
    //	warning("getxy(): No ctype keywords specified or not implemented.");
	return FAILURE;
      }
    }//else
  }
  //  else{
  //    warning("getxy(): division by zero error!");
  //  }

  x0 =  rt * sin( phi )/cdelt1;
  y0 = -rt * cos( phi )/cdelt2;

  // Note that the FORTRAN convention is used for pixel numbers
  // Invert to pixel coordinates
  *x = ipc001001 * x0 + ipc001002 * y0 + crpix1;
  *y = ipc002001 * x0 + ipc002002 * y0 + crpix2;

  return SUCCESS;
}// getxy()


/*****************************************************************************/
//
// FUNCTION    : getPixelsize()
// DESCRIPTION : Return the pixel scales of the image
// 
//
// SIDE EFFECTS:
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
// xscale, yscale       O                    pixel size in x/y direction
//                                           (cdelt1, cdelt2) 
//
//  RETURNS    : SUCCESS
//
/*****************************************************************************/
Status Image::getPixelsize( double *xscale, double *yscale) const
{
  *xscale = cdelt1;
  *yscale = cdelt2;
  return SUCCESS;
}//end getPixelsize( )
