//----------------------------------------------------------------
// srclist.cc
//
// class Srclist implementation
//----------------------------------------------------------------
#include "srclist.h"
//#include "slatec.h"

/*****************************************************************************/
//
// FUNCTION    : Sourcelist() constructor
// DESCRIPTION : 
// 
//
// SIDE EFFECTS: 
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//  n                  I                     maximum number of sources
//
//  RETURNS    : 
//
/*****************************************************************************/
Sourcelist::Sourcelist( unsigned int n)
{
  nsrcs  = 0;
  ndetect= 0;
  nmax = 0;

  allocate( n );

}// Sourcelist() constructor

/*****************************************************************************/
//
// FUNCTION    : Sourcelist() destructor
// DESCRIPTION : 
// 
//
// SIDE EFFECTS: 
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//
//
//  RETURNS    : 
//
/*****************************************************************************/
Sourcelist::~Sourcelist( )
{

  deallocate();

}// ~Sourcelist()

/*****************************************************************************/
//
// FUNCTION    : init()
// DESCRIPTION : Sets number of sources to zero.
// 
//
// SIDE EFFECTS: 
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//
//
//  RETURNS    :
//
/*****************************************************************************/
void Sourcelist::init()
{
  ndetect = 0;
  nsrcs = 0;

}//init()

/*****************************************************************************/
//
// FUNCTION    : allocate()
// DESCRIPTION : Allocates memory for all the arrays of the Sourcelist class.
// 
//
// SIDE EFFECTS: 
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//      n               I              array size = maximum number of sources.
//
//  RETURNS    : SUCCESS or FAILURE;
//
/*****************************************************************************/
Status Sourcelist::allocate(int n)
{
  if( n <= 0 ){
    return FAILURE;
  }//if

  if( nmax > 0 )
    deallocate();

  if( n > 0 )
    nmax = n;
  else
    nmax= NMAX_SRCS;

  //  srcnum = new long [nmax];
  ipix = new unsigned int [nmax];
  jpix = new unsigned int [nmax];
  xima = new float [nmax];
  yima = new float [nmax];
  xerr = new float [nmax];
  yerr = new float [nmax];
  xycorr = new float [nmax];
  scale = new float [nmax];
  sigma = new float [nmax];
  sigmaErr = new float [nmax];
  corr  = new float [nmax];
  corr0 = new float [nmax];
  correrr = new float [nmax];
  crcorr  = new float [nmax];
  cnts  = new float [nmax];
  cntserr = new float [nmax];
  //  extrcnts  = new float [nmax];
  //  extrcntsErr = new float [nmax];
  cntrate = new float [nmax];
  cntraterr  = new float [nmax];
  exposure   = new float [nmax];
  background = new float [nmax];

  radec_err  = new float [nmax];
  ra  = new double [nmax];
  dec = new double [nmax];
  lii = new double [nmax];
  bii = new double [nmax];
  ignore = new bool [nmax];

  //  neighbourNo   = new int [nmax];
  //  neighbourDist = new float [nmax]
  for(unsigned int i=0; i < nmax; i++){
    ignore[i] = false;
    //    neighbourNo[i]   = -1;
    //    neighbourDist[i] = -1.0;
  }//for

  return SUCCESS;
}//allocate()


/*****************************************************************************/
//
// FUNCTION    : deallocate()
// DESCRIPTION : Deallocates the memory of the array elements of class
//               Sourcelist.
// 
//
// SIDE EFFECTS: 
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//
//
//  RETURNS    : SUCCESS or FAILURE
//
/*****************************************************************************/
Status Sourcelist::deallocate( void )
{
  if( nmax > 0 ){
    //    delete [] srcnum;
    delete [] ipix;
    delete [] jpix; 
    delete [] xima; 
    delete [] yima; 
    delete [] xerr;
    delete [] yerr;
    delete [] xycorr;
    delete [] scale;
    delete [] sigma;
    delete [] sigmaErr;
    delete [] corr;
    delete [] corr0;
    delete [] correrr;
    delete [] crcorr;
    delete [] cnts;
    delete [] cntserr;
    //    delete [] extrcnts;
    //    delete [] extrcntsErr;
    delete [] cntrate;
    delete [] cntraterr;
    delete [] exposure;
    delete [] background;
    delete [] radec_err;
    delete [] ra;
    delete [] dec;
    delete [] lii;
    delete [] bii;
    delete [] ignore;
    //    delete [] neighbourNo;
    //    delete [] neighbourDist;
    nmax  = 0;
    nsrcs = 0;
    ndetect = 0;
  }//if

  return SUCCESS;
}//deallocate()

/*****************************************************************************/
//
// FUNCTION    : copy()
// DESCRIPTION : Copies the sources to another source list. Sources with
//               "ignore==true" are not copied.
// 
//
// SIDE EFFECTS: 
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//     srclist          I                       the list to be copied.
//
//  
//  RETURNS    : SUCCESS or FAILURE
//
/*****************************************************************************/
Status Sourcelist::copy( const Sourcelist &srclist)
{
  if( srclist.ndetect == 0 )
    return FAILURE;

  if( srclist.ndetect > nmax ){
    if( allocate( srclist.ndetect ) == FAILURE )
      return FAILURE;
  }//if
  nsrcs = 0;
  for( unsigned int i = 0 ; i < srclist.ndetect; i++){
    if( srclist.ignore[i] == false ){
      //      srcnum[nsrcs] = nsrcs+1;
      ipix[nsrcs] = srclist.ipix[i];
      jpix[nsrcs] = srclist.jpix[i];
      xima[nsrcs] = srclist.xima[i];
      yima[nsrcs] = srclist.yima[i];
      xerr[nsrcs] = srclist.xerr[i];
      yerr[nsrcs] = srclist.yerr[i];
      xycorr[nsrcs] = srclist.xycorr[i];
      scale[nsrcs] = srclist.scale[i];
      sigma[nsrcs] = srclist.sigma[i];
      sigmaErr[nsrcs] = srclist.sigmaErr[i];
      corr[nsrcs]  = srclist.corr[i];
      corr0[nsrcs] = srclist.corr0[i];
      correrr[nsrcs] = srclist.correrr[i];
      crcorr[nsrcs]  = srclist.crcorr[i];
      cnts[nsrcs]    = srclist.cnts[i];
      cntserr[nsrcs] = srclist.cntserr[i];
      //  extrcnts[nsrcs]= srclist.extrcnts[i];
      //      extrcntsErr[nsrcs] = srclist.extrcntsErr[i];
      cntrate[nsrcs]   = srclist.cntrate[i];
      cntraterr[nsrcs] = srclist.cntraterr[i];
      exposure[nsrcs]  = srclist.exposure[i];
      background[nsrcs]= srclist.background[i];
      radec_err[nsrcs] = srclist.radec_err[i];
      ra[nsrcs]  = srclist.ra[i];
      dec[nsrcs] = srclist.dec[i];
      lii[nsrcs] = srclist.lii[i];
      bii[nsrcs] = srclist.bii[i];
      ignore[nsrcs] = srclist.ignore[i];
      //      neighbourNo[nsrcs]  = srclist.neighbourNo[i];
      // neighbourDist[nsrcs] = srclist.neighbourDist[i];
      nsrcs++;
    }//if
  }//for

  ndetect = nsrcs;

  //  copy_keywords(srclist);

  return SUCCESS;
}// copy()

/*****************************************************************************/
//
// FUNCTION    : reducelist()
// DESCRIPTION : Checks whether a source is found at several wavelet scales.
//               This is done by comparing a given source with sources detected
//               at another waveletscale. Only one per wavelet scale can be 
//               deleted. To do so for each wavelet scale the source closest
//               to source i will be found and its correlation value compared
//               to that of source i. The one with the smallest correlation 
//               value is then tagged ignore=true.
//
//               In order to avoid confusion at larger wavelet scales, the
//               selection is done in two stages:
//
//               I 
//               Compare sources close to each other at different wavelet 
//               scales. If corr1 > corr2 (scale1 < scale1) ignore source 2
//
//               II
//               Compare sources close to each other at different wavelet 
//               scales. If corr1 > corr2 (scale1 < scale1) ignore source 2,
//               otherwise ignore source 1.
//
//
//               It assumes the sources are ordered on scale size.
//
// SIDE EFFECTS:
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//
//
//  RETURNS    : SUCCESS or FAILURE
//
/*****************************************************************************/
#define  WMINSEP 2.240926
#define SCALETOL 0.1  // scale tolerance, based on the fact that 
                      // (0.5-sqrt(0.5)*0.5)= 0.15. Usually scale >= 1, 
                      // so 0.1 is safe.
Status Sourcelist::reducelist( void )
{

  if( ndetect <= 0 )
    return FAILURE;

  nsrcs=0;
  for( unsigned int i = 0; i < ndetect; i++ )
    if( ignore[i] == false) nsrcs++;

  //  printf("reducelist(): nsrcs = %i\n", nsrcs);


  int jmin, istart=0;
  float compscale, mindist, dist, scalesize,maxscale =  scale[ndetect-1];

  for( unsigned int i = 0; i < ndetect-1 ; i++ ){

    if( ignore[i] == false ){

      compscale = scale[i]*M_SQRT2;
      while( compscale < 1.1*maxscale ){
	jmin = -1;
	mindist = MAX_FLOAT;
	for(  unsigned int j = i+1; j < ndetect ; j++ ){
	  dist = sqrt(  SQR(xima[i]-xima[j]) + SQR(yima[i]-yima[j]));
	  if( ignore[j] == false && dist <= scale[j] ){
	    
	    if( dist < mindist ){
	      mindist = dist;
	      jmin = j;
	    }//if
	  }//if
	}//for j

	if( jmin != -1 ){
	  if( corr[i] >= corr[jmin] )
	    ignore[jmin] = true;
	}//if

	compscale *= M_SQRT2;
      }//while
    }//if
      
  }//for i


  compscale = scale[0]*M_SQRT2;
  while( compscale <= maxscale*1.1 ){ 
  
    istart = 0;
    scalesize = scale[0];
    while( scalesize < 0.9*maxscale){

      for( unsigned int i = istart ; 
	   i < ndetect-1 && scale[i]<=1.1*scalesize; i++ ){
	if( ignore[i] == false && fabs(scale[i]-scalesize) < SCALETOL){

	  
	  jmin = -1;
	  mindist = MAX_FLOAT;
	  for(  unsigned int j = i+1; 
		j < ndetect && scale[j] < 0.9*compscale*M_SQRT2; j++ ){
	    dist = sqrt(  SQR(xima[i]-xima[j]) + SQR(yima[i]-yima[j]));
	    if( ignore[j] == false && dist <= scale[j] &&
		fabs(scale[j]-compscale) < SCALETOL ){
	      if( dist < mindist ){
		mindist = dist;
		jmin = j;
	      }//if
	    }//if
	  }//for j

	  if( jmin != -1 ){
	    if( corr[i] >= corr[jmin] )
	      ignore[jmin] = true;
	    else
	      ignore[i] = true;
	  }//if
	  
	}//if
	
	istart++;    
      }//for i
      
      scalesize *= M_SQRT2;
    }//while
    compscale *= M_SQRT2;
  }//while

  nsrcs=0;
  for( unsigned int i = 0; i < ndetect; i++ )
    if( ignore[i] == false) nsrcs++;

  //  printf("reducelist(): nsrcs = %i\n", nsrcs);
  
  return SUCCESS;

}// reducelist()


/*****************************************************************************/
//
// FUNCTION    : smallfirst()
// DESCRIPTION : Compares a given wavelet correlation (=amplitude) with
//               amplitudes at smaller wavelet scales.
//               In particular it looks whether the candidate may actually
//               consists of multiple sources within the boundary of the
//                candidate. If so the candidate is ignored.
// 
//
// SIDE EFFECTS: Uses the fact that the list is ordered from small to large
//               wavelet scales.
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//
//
//  RETURNS    : SUCCESS or FAILURE
//
/*****************************************************************************/
Status Sourcelist::smallfirst( void )
{

  if( ndetect <= 0 || reducelist()==FAILURE )
    return FAILURE;

  float xdist, ydist, maxcorr,totcorr, totcnts, compscale, minscale=scale[0];
  unsigned int i, j, ncontained;

  for( i = ndetect-1; i > 1; i--){

    if( ignore[i] == false ){
      compscale = scale[i]/M_SQRT2;
      while( compscale >= 0.9*minscale ){

	ncontained = 0;
	totcorr = 0.0;
	totcnts= 0.0;
	maxcorr = 0.0;
	for( j = i-1; j > 0 && scale[j] >= 0.4*compscale ; j-- ){
	 
	  if( ignore[j] == false && scale[j] <= compscale){

	    xdist=  xima[i] - xima[j];
	    ydist=  yima[i] - yima[j];
	    
	    if(sqrt(xdist*xdist + ydist*ydist) <= 2*scale[i] ){
	      totcnts += cnts[j];
	      totcorr += corr[j];
	      ncontained++;
	      if( corr[j] > maxcorr)
		maxcorr = corr[j];
	    }
	  }//if
	}//for j
	if( ncontained >= 2 && maxcorr > corr[i]){
	  ignore[i] = true;
	}
	else if( ncontained >= 2){
	  // go again through the list and throw out the sources
	  for( j = i-1; j > 0 ; j-- ){
	    
	    if( ignore[j] == false ){
	      xdist=  xima[i] - xima[j];
	      ydist=  yima[i] - yima[j];
	      if(sqrt(xdist*xdist + ydist*ydist) <= 2*scale[i] ){
		ignore[j] = true;
	      }
	    }//if
	  }//for j
	  
	}//else if

	compscale/= M_SQRT2;
      }//while
    }//if( ignore
  }//for i

  nsrcs=0;
  for( i = 0; i < ndetect; i++ )
    if( ignore[i] == false) nsrcs++;

  return SUCCESS;

}// smallfirst()


/*****************************************************************************/
//
// FUNCTION    : calcScale() 
// DESCRIPTION : Uses the detection of a source in several wavelet scales to
//               improve on the measurement of the source extent (sigma).
//               Only scales close to the optimum scale are used. Errors are
//               calculated using some analytical approximation, whihc are 
//               valid in case the the source has a gaussian shape.
//               The correlation between correlation values at different
//               scales are not yet taken into account (see Damiani et al.
//               for a possible approach).
// 
//
// SIDE EFFECTS: 
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//
//
//  RETURNS    : SUCCESS or FAILURE
//
/*****************************************************************************/
#define MINNOM 1.0e-2
Status Sourcelist::calcScale()
{

  if( nsrcs == 0 ){
    //    Sas::warning("noSources",
    //		 "No sources detected sofar, no further processing.");
    //    warning("No sources detected sofar, no further processing.");
    return FAILURE;
  }//if


  float mindist, dist;
  float sigma1, sigma2;
  float corr1err, corr2err, cnts1, cnts2, cnts1err ,cnts2err;
  //  float cntr1, cntr2, cntr1err, cntr2err;
  float sigma1err, sigma2err;
  
  unsigned int j, jmin=0;
  for( unsigned int i=0; i < ndetect ; i++){
    
    corr1err = corr2err = cnts1err= cnts2err = 0.0;

    if( ignore[i] == false ){
      sigma1 = -1.0; // initialize
      sigma2 = -1.0;
      sigmaErr[i] = 0.0;

      //
      // First look to scales smaller than best source scale
      //
      mindist= MAX_FLOAT;
      jmin = MAX_INT;
      for( j = 0 ; j < ndetect ; j++){
	if(j != i && scale[j] < 0.9*scale[i] && scale[j] > 0.5*scale[i] ){
	  dist =sqrt( SQR(xima[j]-xima[i]) + SQR(yima[j]-yima[i]));
	  if( dist < mindist && dist < scale[j] ){
	    mindist = dist;
	    jmin = j;
	  }//if
	}//if
      }//for

      if( jmin != MAX_INT && corr[jmin] > 0.0){
	calcProperties(i, jmin, &sigma1, &cnts1, &corr1err, 
		       &cnts1err, &sigma1err);
      }//if

      //
      // Then look to scales larger than best source scale
      //
      mindist= MAX_FLOAT;
      jmin = MAX_INT;
      for( j = i+1 ; j < ndetect ; j++){
	if(j != i && scale[j] > 1.1*scale[i] && scale[j] < 1.9*scale[i] ){
	  dist =sqrt( SQR(xima[j]-xima[i]) + SQR(yima[j]-yima[i]));
	  if( dist < mindist && dist < scale[i] ){
	    mindist = dist;
	    jmin = j;
	  }//if
	}//if
      }//for


      if( jmin != MAX_INT && corr[jmin] > 0.0){
	calcProperties(i, jmin, &sigma2, &cnts2, &corr2err, 
		       &cnts2err, &sigma2err);
      }//if


   if( sigma2 <= 0.0 && sigma1 <= 0.0 ){
	  sigma[i] = scale[i];
	  sigmaErr[i] = (M_SQRT2-1.0) * scale[i];
	  correrr[i]=correlationError(scale[i],sigma[i],background[i],cnts[i]);
	  cntserr[i] = correrr[i]* 4.0 * M_PI * SQR(scale[i]);
      }
      else 
	if( (sigma1 > 0.0 && sigma2 > 0.0 && sigma1err < sigma2err) ||
	    sigma1 > 0.0 && sigma2 < 0.0 ){
	  // choose the best since averaging is wrong as sigma1&sigma2 are 
	  // statistically dependent
	  sigma[i]= sigma1;
	  cnts[i] = cnts1;
	  cntrate[i] = cnts[i]/exposure[i];
	  correrr[i] = corr1err;
	  cntserr[i] = cnts1err;
	  cntraterr[i] = cntserr[i]/exposure[i];	    
	  sigmaErr[i]  = sigma1err;
	}
	else 
	  if( sigma2 > 0.0 ){ //  && sigma1err > sigma2er
	    sigma[i]= sigma2;
	    cnts[i] = cnts2;
	    cntrate[i] = cnts[i]/exposure[i];
	    correrr[i] = corr2err;
	    cntserr[i] = cnts2err;
	    cntraterr[i] = cntserr[i]/exposure[i];	    
	    sigmaErr[i]  = sigma2err;
	  }
   if (exposure[i] > 0.0 ){
      cntrate[i] = cnts[i]/exposure[i];
      cntraterr[i] = cntserr[i]/exposure[i]; 
   }
   else {
     cntrate[i]=0.0;
     cntraterr[i]=0.0;
   } 
    }//if( ignore[i] ==

  }//for i

  return SUCCESS;
}//calcScale

/*****************************************************************************/
//
// FUNCTION    : calcProperties
// DESCRIPTION : Calculates the "real" sigma of the source and the errors of
//               various source parameters. The input consists of the index
//                numbers associated with the same source at two different
//                wavelet scales.
//
//
//
//
// SIDE EFFECTS: 
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
// i, jmin              I                      index number of two detections
//                                             of same source
// sigma2               O                      extent estimate
// cnts2                O                      counts estimate
// cntr2                O                      countrate estimate
// corr2err             O                      error in correlation value
// cnts2err             O                      error in cnts2
// cntr2err             O                      error in cntr2
// sigma2err            O                      error in sigma2
//
//  RETURNS    : SUCCESS or FAILURE
//
/*****************************************************************************/
Status Sourcelist::calcProperties(unsigned int i, unsigned int jmin,
				  float *sigma2, float *cnts2, 
				  float *corr2err, 
				  float *cnts2err,
				  float *sigma2err)
{

  float sqcratio, scrat, sigsqr, corr2cnts, corr3err, ratioError, cov;


  sqcratio = sqrt(corr[i]/corr[jmin]);
  scrat = scale[i]/scale[jmin];
  if( fabs(sqcratio - scrat) > MINNOM ){
    sigsqr = (scrat*SQR(scale[jmin]) - sqcratio*SQR(scale[i]))/
      (sqcratio - scrat);
    if( sigsqr > 0.0 ){
      *sigma2  = sqrt( sigsqr );
      corr2cnts = M_PI*SQR( SQR( *sigma2)+SQR(scale[i]))/SQR(scale[i]);
      *cnts2 = corr[i]*corr2cnts;
       //      *cntr2 = crcorr[i]*corr2cnts;
      //      *cntr2 = *cnts2/exposure[i];
      corr3err= correlationError(scale[jmin], *sigma2,background[i], *cnts2);
      *corr2err = correlationError(scale[i], *sigma2, background[i], *cnts2);
      *cnts2err = *corr2err * corr2cnts; //+ additional terms
      //      *cntr2err = *cnts2err/exposure[i];
      cov = correlationCovariance( scale[i],scale[jmin], *sigma2,
				   background[i], *cnts2);

      ratioError = correlationRatioError(corr[i], *corr2err, 
					 corr[jmin],corr3err, cov);
      *sigma2err = sigmaError( sqcratio, ratioError, scale[i],scale[jmin],
			      *sigma2 );
    }//if
  }
  else
    return FAILURE;

  
    return SUCCESS;
}//calcProperties()


/*****************************************************************************/
//
// FUNCTION    : statistics()
// DESCRIPTION : Calculates average #counts and its variance.
//               Used for debugging/checking the code
//               
// 
//
// SIDE EFFECTS: 
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//
//
//  RETURNS    : SUCCESS or FAILURE
//
/*****************************************************************************/
Status Sourcelist::statistics()
{


  if( ndetect <= 2 )
    return FAILURE;

  nsrcs = 0;
  float sum=0.0;
  unsigned int i;
  for( i = 0; i < ndetect; i++ ){

    if( ignore[i] == false ){
      nsrcs++;
      sum += cnts[i];
    }//if

  }//for i

  float mean = sum/nsrcs;
  cout << "#sources = "<< nsrcs << endl;
  cout << "average cnts = "<< mean << endl;

  float var=0.0;
  for( i = 0; i < ndetect; i++ ){

    if( ignore[i] == false ){
      var += SQR((cnts[i]-mean)/cntserr[i]);
    }//if

  }//for i

  var/= nsrcs;

  cout << "variance = "<< var << endl;
  cout << "std deviation = "<< sqrt(var*nsrcs/(nsrcs-1)) << endl;

  return SUCCESS;

}//

/*****************************************************************************/
//
// FUNCTION    : calcCoordinates()
// DESCRIPTION : Fills in the coordinates ra and dec using an Image object.
//               
// 
//
// SIDE EFFECTS: 
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//      image          I             Image with a world coordinate system
//
//  RETURNS    : SUCCESS or FAILURE
//
/*****************************************************************************/
Status Sourcelist::calcCoordinates(const Image &im)
{
  if( ndetect == 0 ){
    return FAILURE;
  }//if

  double cdelt1, cdelt2;
  /*  extern "FORTRAN" void sla_EQGAL(double ra, double dec, double *lii, double *bii){
  void sla_eqgal(); };
  */

  im.getPixelsize( &cdelt1, &cdelt2 );
  if( cdelt1 < 0.0 )
    cdelt1 *= -1.0;
  if( cdelt2 < 0.0 )
    cdelt2 *= -1.0;

  for(unsigned int i = 0; i < ndetect ; i++){
    if( im.getradec(xima[i], yima[i], &ra[i], &dec[i] ) == SUCCESS ){
      radec_err[i] = 3600.0*sqrt(0.5*(SQR( (float)cdelt1 * xerr[i]) + 
				      SQR( (float)cdelt2 * yerr[i])));
      lii[i] = bii[i] = 0.0;
      
      //  sla_EQGAL ( ra[i], dec[i], &lii[i], &bii[i]);
    }
    else{
      ra[i] = dec[i] = 0.0;
      Sas::warning("coordError","Error in ra/dec calculation encountered!");
    }
  }//for


  return SUCCESS;
}//

/*****************************************************************************/
//
// FUNCTION    : copyKeyword()
// DESCRIPTION : Copies the keywords for the primary header from an 
//               image object
//               
// 
//
// SIDE EFFECTS: 
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//      image          I             Image with a world coordinate system
//
//  RETURNS    : SUCCESS or FAILURE
//
/*****************************************************************************/
void Sourcelist::copyKeywords(const Image &im)
{

  instrument = im.instrument;
  telescope  = im.telescope;
  detnam     = im.detnam;
  datamode   = im.datamode;
  filter     = im.filter;
  obs_mode   = im.obs_mode;
  obs_id     = im.obs_id;
  exp_id     = im.exp_id;
  object     = im.object;
  obsdate    = im.obsdate;
  enddate    = im.enddate;
  radecsys   = im.pradecsys;
  equinox   = im.pequinox;
  pexposure  = im.pexposure;
  ra_pnt    = im.ra_pnt;
  dec_pnt   = im.dec_pnt;
  pa_pnt    = im.pa_pnt;

}// copyKeywords()

/*****************************************************************************/
//
// FUNCTION    : readset()
// DESCRIPTION : Read a fits source list file.
// 
//
// SIDE EFFECTS: 
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//   setname            I                    Name of the dataset to be written
//
//  RETURNS    : SUCCESS or FAILURE
//
/*****************************************************************************/
Status Sourcelist::readset( const string &setname )
{ 

  float fwhm = sqrt( 8.0 * log(2.0));
  
  DataSet *srclistSet = dataSetServer->open(setname,Dal::Read);
  Table *table = srclistSet->table("SRCLIST");

  unsigned int nrows = table->rows();

  ndetect = nrows;
  nsrcs = nrows; 


  //  ColumnData *cdnum   = table->column("SRC_NUM")->data();
  ColumnData *cdimax  = table->column("IMAX")->data(); 
  ColumnData *cdjmax  = table->column("JMAX")->data();
  ColumnData *cdxima = table->column("X_IMA")->data();
  ColumnData *cdyima = table->column("Y_IMA")->data(); 
  ColumnData *cdxerr  = table->column("X_IMA_ERR")->data();
  ColumnData *cdyerr  = table->column("Y_IMA_ERR")->data();
  ColumnData *cdra    = table->column("RA")->data();
  ColumnData *cddec   = table->column("DEC")->data();
  ColumnData *cdradec_err = table->column("RADEC_ERR")->data();
  ColumnData *cdscnts  = table->column("SCNTS")->data();
  ColumnData *cdscntserr = table->column("SCNTS_ERR")->data();
  ColumnData *cdcorr  = table->column("WCORR")->data();
  ColumnData *cdscale = table->column("WSCALE")->data(); 
  ColumnData *cdext = table->column("EXTENT")->data();
  ColumnData *cdext_err = table->column("EXT_ERR")->data(); 
  ColumnData *cdrate = table->column("RATE")->data(); 
  ColumnData *cdrate_err = table->column("RATE_ERR")->data(); 
  ColumnData *cdexp_map = table->column("EXP_MAP")->data(); 
  ColumnData *cdbg_map = table->column("BG_MAP")->data(); 
  ColumnData *cdxycorr = table->column("XYCORR")->data(); 


  int *pipix = cdimax->int32Data();
  int *pjpix = cdjmax->int32Data();
  real32 *pxima    = cdxima->real32Data();
  real32 *pyima    = cdyima->real32Data();
  real32 *pxerr    = cdxerr->real32Data();
  real32 *pyerr    = cdyerr->real32Data();
  real32 *pradec_err = cdradec_err->real32Data();
  real32 *pcnts    = cdscnts->real32Data();
  real32 *pcntserr = cdscntserr->real32Data();
  real32 *pcorr    = cdcorr->real32Data();
  real32 *pscale   = cdscale->real32Data();
  real32 *psigma   = cdext->real32Data();
  real32 *psigmaErr   = cdext_err->real32Data();
  real32 *pcntrate = cdrate->real32Data();
  real32 *pcntraterr = cdrate_err->real32Data();
  real32 *pexposure = cdexp_map->real32Data();
  real32 *pbackground = cdbg_map->real32Data();
  real32 *pxycorr = cdxycorr->real32Data();  
  real64 *pra = cdra->real64Data();
  real64 *pdec = cddec->real64Data();
   
  for (unsigned int i=0; i<nrows; i++){
     ipix[i] =        pipix[i];
     jpix[i] =        pjpix[i];
     xima[i] =        pxima[i];
     yima[i] =        pyima[i];
     xerr[i] =        pxerr[i];
     yerr[i] =        pyerr[i];
     radec_err[i] =   pradec_err[i];
     cnts[i] =        pcnts[i];
     cntserr[i] =     pcntserr[i];
     corr[i] =        pcorr[i];
     scale[i] =       pscale[i];
     sigma[i] =       psigma[i]/fwhm;
     sigmaErr[i] =    psigmaErr[i]/fwhm;
     cntrate[i] =     pcntrate[i];
     cntraterr[i] =   pcntraterr[i];
     exposure[i] =    pexposure[i];
     background[i] =  pbackground[i];
     xycorr[i] =      pxycorr[i];
     ra[i] =          pra[i];
     dec[i] =         pdec[i];
     ignore[i]  =     false;
  } 

  dataSetServer->close(srclistSet);
   
  delete  cdimax;
  delete  cdjmax;
  delete  cdxima;
  delete  cdyima;
  delete  cdxerr;
  delete  cdyerr;
  delete  cdradec_err;
  delete  cdscnts;
  delete  cdscntserr;
  delete  cdcorr;
  delete  cdscale;
  delete  cdext;
  delete  cdext_err;
  delete  cdrate;
  delete  cdrate_err;
  delete  cdexp_map;
  delete  cdbg_map;
  delete  cdxycorr;
  delete  cdra;
  delete  cddec;

  return SUCCESS;

}// readset()

/*****************************************************************************/
//
// FUNCTION    : writeset()
// DESCRIPTION : Writes the data to a fits file.
// 
//
// SIDE EFFECTS: 
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//   setname            I                    Name of the dataset to be written
//
//  RETURNS    : SUCCESS or FAILURE
//
/*****************************************************************************/
Status Sourcelist::writeset( const string &setname )
{
  nsrcs=0;
  for( unsigned int i = 0; i < ndetect; i++ )
    if( ignore[i] == false) nsrcs++;

 //  if( nsrcs == 0 ){
    // error("No source list to be written!");
 //    return FAILURE;
 //  }//if

  DataSet *srclistSet = dataSetServer->open(setname,Dal::Create);

  *srclistSet->addAttribute("TELESCOP","Telescope (mission) name")= telescope;
  *srclistSet->addAttribute("INSTRUME","Instrument name")= instrument;
  *srclistSet->addAttribute("DETNAM","Prime or redundant")= detnam;
  *srclistSet->addAttribute("DATAMODE","Instr. mode (IMAGING,TIMING,..)")= datamode;
  *srclistSet->addAttribute("FILTER","Filter ID")= filter;
  *srclistSet->addAttribute("OBS_MODE","Obs. mode (pointing or slew)")= obs_mode;
  *srclistSet->addAttribute("OBS_ID","Observation identifier")= obs_id;
  *srclistSet->addAttribute("EXP_ID","Exposure identifier")= exp_id;
  *srclistSet->addAttribute("CONTENT","Content of file")= content;
  //  *srclistSet->addAttribute("ORIGIN","")= origin;

  *srclistSet->addAttribute("OBJECT","Name of observed object") = object;
  *srclistSet->addAttribute("RA_PNT","RA of instr. pointing") = ra_pnt;
  *srclistSet->addAttribute("DEC_PNT","DEC of instr. pointing") = dec_pnt;
  *srclistSet->addAttribute("PA_PNT","RA of instr. pointing") = ra_pnt;

  *srclistSet->addAttribute("RADECSYS","Coordinate system")= radecsys;
  *srclistSet->addAttribute("EXPOSURE","Exposure (sec)") = exposure;
  *srclistSet->addAttribute("DATE-OBS","Start of the observation","D/M/Y") = obsdate;
  *srclistSet->addAttribute("DATE-END","End of the observation","D/M/Y") = enddate;
  *srclistSet->addAttribute("OBSERVER","Principle Investigator") = observer;


  Table *table = srclistSet->addTable("SRCLIST",nsrcs);
  ColumnData *col[30];
  int32 *srcnum, *imax, *jmax;
  float *fdata[30];
  double *ddata[5];


  *table->addAttribute("RADECSYS","Coordinate system")= radecsys;
  *table->addAttribute("EQUINOX","Coordinate system")= equinox;

  table->addColumn("SRC_NUM",Dal::Int32);
  table->addColumn("IMAX",Dal::Int32);
  table->addColumn("JMAX",Dal::Int32);
  table->addColumn("X_IMA",Dal::Real32);
  table->addColumn("Y_IMA",Dal::Real32);
  table->addColumn("X_IMA_ERR",Dal::Real32);
  table->addColumn("Y_IMA_ERR",Dal::Real32);
  table->addColumn("RA",Dal::Real64);
  table->addColumn("DEC",Dal::Real64);
  table->addColumn("RADEC_ERR",Dal::Real32);
  //  table->addColumn("LII",Dal::Real64);
  //  table->addColumn("BII",Dal::Real64);
  table->addColumn("SCNTS",Dal::Real32);
  table->addColumn("SCNTS_ERR",Dal::Real32);
  table->addColumn("WCORR",Dal::Real32);
  table->addColumn("WSCALE",Dal::Real32);
  table->addColumn("EXTENT",Dal::Real32);
  table->addColumn("EXT_ERR",Dal::Real32);
  table->addColumn("RATE",Dal::Real32);
  table->addColumn("RATE_ERR",Dal::Real32);
  table->addColumn("EXP_MAP",Dal::Real32);
  table->addColumn("BG_MAP",Dal::Real32);
  table->addColumn("XYCORR",Dal::Real32);
  //  table->addColumn("THRESHOLD",Dal::Real32);

  unsigned int i=0;
  col[i] = table->column("SRC_NUM")->data();
  srcnum = col[i]->int32Data();
  i++;
  col[i] = table->column("X_IMA")->data();
  fdata[i] = col[i]->real32Data();
  i++;
  col[i] = table->column("X_IMA_ERR")->data();
  fdata[i] = col[i]->real32Data();
  i++;
  col[i] = table->column("Y_IMA")->data();
  fdata[i] = col[i]->real32Data();
  i++;
  col[i] = table->column("Y_IMA_ERR")->data();
  fdata[i] = col[i]->real32Data();
  i++;
  col[i] = table->column("SCNTS")->data();
  fdata[i] = col[i]->real32Data();
  i++;
  col[i] = table->column("SCNTS_ERR")->data();
  fdata[i] = col[i]->real32Data();
  i++;
  col[i] = table->column("WSCALE")->data();
  fdata[i] = col[i]->real32Data();
  i++;
  col[i] = table->column("XYCORR")->data();
  fdata[i] = col[i]->real32Data();
  i++;
  col[i] = table->column("RATE")->data();
  fdata[i] = col[i]->real32Data();
  i++;
  col[i] = table->column("RATE_ERR")->data();
  fdata[i] = col[i]->real32Data();
  i++;
  col[i] = table->column("WCORR")->data();
  fdata[i] = col[i]->real32Data();
  i++;
  col[i] = table->column("EXP_MAP")->data();
  fdata[i] = col[i]->real32Data();
  i++;
  col[i] = table->column("BG_MAP")->data();
  fdata[i] = col[i]->real32Data();
  i++;
  col[i] = table->column("RADEC_ERR")->data();
  fdata[i] = col[i]->real32Data();
  i++;
  col[i] = table->column("EXTENT")->data();
  fdata[i] = col[i]->real32Data();
  i++;
  col[i] = table->column("EXT_ERR")->data();
  fdata[i] = col[i]->real32Data();
  i++;
  //  col[i] = table->column("THRESHOLD")->data();
  //  fdata[i] = col[i]->real32Data();
  //  i++;


  col[i] = table->column("RA")->data();
  ddata[0] = col[i]->real64Data();
  i++;
  col[i] = table->column("DEC")->data();
  ddata[1] = col[i]->real64Data();
  i++;
  //  col[i] = table->column("LII")->data();
  //  ddata[2] = col[i]->real64Data();
  //  i++;
  //  col[i] = table->column("BII")->data();
  //  ddata[3] = col[i]->real64Data();
  //  i++;
  col[i] = table->column("IMAX")->data();
  imax = col[i]->int32Data();
  i++;
  col[i] = table->column("JMAX")->data();
  jmax = col[i]->int32Data();
  i++;


  unsigned int n=0;
  float fwhm= sqrt( 8.0 * log(2.0));
  for( unsigned int j=0 ; j < ndetect ; j++){
    
    if( ignore[j] ==false ){
      srcnum[n]= n+1;
      fdata[1][n] = xima[j];
      fdata[2][n] = xerr[j];
      fdata[3][n] = yima[j];
      fdata[4][n] = yerr[j];
      fdata[5][n] = cnts[j];
      fdata[6][n] = cntserr[j];
      fdata[7][n] = scale[j];
      fdata[8][n] = xycorr[j];
      fdata[9][n] = cntrate[j];
      fdata[10][n] = cntraterr[j];
      fdata[11][n] = corr[j];
      fdata[12][n] = exposure[j];
      fdata[13][n] = background[j];
      fdata[14][n] = radec_err[j];
      fdata[15][n] = fwhm*sigma[j];
      fdata[16][n] = fwhm*sigmaErr[j];
      //      fdata[16][n] = corr0[j];

      imax[n] = ipix[j];
      jmax[n] = jpix[j];
      ddata[0][n] = ra[j];
      ddata[1][n] = dec[j];
      //  ddata[2][n] = lii[j];
      //  ddata[3][n] = bii[j];
      n++;
    }
  }

  for( unsigned int k = 0; k < i ; k++)
    delete col[k];


  dataSetServer->close(srclistSet);

  return SUCCESS;
}//writeset()
 


/*****************************************************************************/
//
// FUNCTION    :  correlationCovariance()
// DESCRIPTION :  Calculates the covariance of the wavlet correlation, assuming
//                a gaussian source shape.
// 
//
// REMARKS     :
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//    scale1,scale2     I                      wavelet scales
//    sigma             I                      inferred extend of source
//    bgd               I                      background photons/pixel
//    nrsc              I                      source normalization 
//
//  RETURNS    :
//
/*****************************************************************************/
float correlationCovariance(float scale1, float scale2, float sigma, float bgd,
			    float nsrc)
{

  float covariance, ssqr, t;

  t = SQR(scale1*scale2) + SQR(scale1*sigma) + SQR(scale2*sigma);
  ssqr = 2.0*SQR(sigma*scale1*scale2)/t;

  covariance = nsrc/(2.0 * SQR(M_PI) * t) * 
    (2.0-ssqr/SQR(scale1)-ssqr/SQR(scale2)+SQR(ssqr)/SQR(scale1*scale2) ) +
    4.0*SQR(scale1)*SQR(scale2)*bgd/(M_PI*pow(SQR(scale1)+SQR(scale2),3.0));
  
  return covariance;
  
}//correlationCovariance()

/*****************************************************************************/
//
// FUNCTION    :  correlationError()
// DESCRIPTION :  Calculates the error in the correlation value using
//                an anylytical function which assumes the source has a 
//                gaussian shap and a constant background.
// 
//
// REMARKS     :
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//    scale             I                      wavelet scale
//    sigma             I                      inferred extend of source
//    bgd               I                      background photons/pixel
//    nrsc              I                      source normalization 
//
//  RETURNS    :
//
/*****************************************************************************/
float correlationError( float scale, float sigma, float bgd, float nsrc)
{

  return sqrt(correlationCovariance(scale, scale, sigma, bgd, nsrc) );
}// correlationError()



/*****************************************************************************/
//
// FUNCTION    :  correlationRatioError()
// DESCRIPTION :  Calculates the error in C1/C2, the ratio of correlations.
//                Note the use of the covariance as C1 & C2 are not stat.
//                independent. (cf Damiani et al 1997).
// 
//
// REMARKS     :
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//      corr1           I                    C1
//      dcorr1          I                    error in C1
//      corr2           I                    C2
//      dcorr2          I                    error in C2
//
//  RETURNS    : the error in C1/C2
//
/*****************************************************************************/
float correlationRatioError(float corr1, float dcorr1, 
			    float corr2, float dcorr2, float covariance)
{
  float err2, coeff1 = 1.0/corr2, coeff2 = -corr1/SQR(corr2);

  err2 = SQR(coeff1*dcorr1) + SQR(coeff2*dcorr2) + 2*coeff1*coeff2*covariance;

  return sqrt(err2);
}//correlationRatioError()

/*****************************************************************************/
//
// FUNCTION    :  sigmaError()
// DESCRIPTION : Calculates the error in sigma (extent of gaussian source).
//               given the square root of the correlation ratio sqrt(C1/C2)
//               and the two scale size. It also needs the error in C1/C2.
// 
//
// REMARKS     :
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
// sqrtratio            I                      sqrt(C1/C2) 
// corrRatioError       I                      error of C1/C2
// scale1,scale2        I                      wavelet scales
// sigma                I                      source extent in gaussian sigmas
//
//  RETURNS    : the error of sigma
//
/*****************************************************************************/
float sigmaError( float sqrtratio, float corrRatioError, float scale1,
		  float scale2, float sigma)
{

  float error;
  
  error = 0.25*scale1/(sqrtratio*sigma) * (SQR(scale1)/scale2 - scale2)/
    SQR(sqrtratio - scale1/scale2) *corrRatioError;

  return fabs(error);
}// sigmaError()
