#include <stdio.h>
#include <math.h>
#include <string.h>
#include "Dal.h"
#include "Error.h"
#include "errstr.h"

#define MAXDATA 100

int main( int argc, char *argv[] )
{

  if( argc != 2)
    Sas::error("paramError","One commandline parameter should be specified");

  unsigned long ndata=MAXDATA; 
  float x[MAXDATA], y[MAXDATA], cnts[MAXDATA], wscale[MAXDATA],extent[MAXDATA];
  float radec_err[MAXDATA];
  double ra[MAXDATA], dec[MAXDATA];  
  int readsrclist(string filename, unsigned long *ndata, 
		  float *x, float *y, float *cnts, float *wscale,float *extent,
		   double *ra, double *dec, float *radec_err   );



  if( !readsrclist( argv[1], &ndata, x, y,cnts, wscale, extent,ra, dec, 
		    radec_err ) ){
    errstr << "Error reading " << argv[1] << Sas::error("readError");
  }

  if( ndata != 57 )
    errstr << "Incorrect number of sources: " << ndata << 
      Sas::error("nsrcError");

  int testrow = 28;
  if(fabs( x[testrow]-1.022843E+02) > 0.01
     || fabs(y[testrow]-1.257932E+02) > 0.01 ){
    //    printf("%f %f\n",x[testrow],y[testrow]);
    errstr << "Error in position source #" << testrow+1 << 
      Sas::error("possError");
  }
  if( fabs( ra[testrow] - 10.729064940) > 2.0/3600.0 || 
      fabs( dec[testrow]- 41.267741166) > 2.0/3600.0 )
    errstr << "Error in coordinates source #" << testrow+1 << 
      Sas::error("coordError");

  if( fabs( radec_err[testrow] - 1.039559E-01) > 0.001 )
      errstr << "Error in position error source #" << testrow+1 << 
	Sas::error("poserrError");

  //  cout << cnts[testrow] << endl;

  if( fabs(cnts[testrow]-636.6840) > 0.1 )
    errstr << "Error in count estimate source #" << testrow+1 << 
      Sas::error("countError");

  if( fabs(wscale[testrow]-2.0)>0.01 || fabs(extent[testrow]-2.786)>0.01)
    errstr << "Error in wavelet scale or extent source #" << testrow+1 << 
      Sas::error("scaleError");


  return 0;
}//int main()



/*
  Read some columns from the input file


 */
int readsrclist( string setname, unsigned long *ndata, 
		 float *x, float *y, float *cnts, float *wscale, float *extent,
		 double *ra, double *dec, float *radec_err   )
{

  DataSet *srclistSet = dataSetServer->open(setname,Dal::Read);
  Table *table = srclistSet->table("SRCLIST");
  ColumnData *col[30];
  float *fdata[30];
  double *ddata[3];


  *ndata= (*ndata > table->rows() ? table->rows() : *ndata );

  unsigned int i=0;
  col[i] = table->column("X_IMA")->data();
  fdata[i] = col[i]->real32Data();
  i++;
  col[i] = table->column("Y_IMA")->data();
  fdata[i] = col[i]->real32Data();
  i++;
  col[i] = table->column("SCNTS")->data();
  fdata[i] = col[i]->real32Data();
  i++;
  col[i] = table->column("WSCALE")->data();
  fdata[i] = col[i]->real32Data();
  i++;
  col[i] = table->column("EXTENT")->data();
  fdata[i] = col[i]->real32Data();
  i++;
  col[i] = table->column("RADEC_ERR")->data();
  fdata[i] = col[i]->real32Data();
  i++;
  col[i] = table->column("RA")->data();
  ddata[0] = col[i]->real64Data();
  i++;
  col[i] = table->column("DEC")->data();
  ddata[1] = col[i]->real64Data();
  i++;


  for( unsigned int j=0 ; j < *ndata  ; j++){

    x[j] = fdata[0][j];
    y[j] = fdata[1][j];
    cnts[j]   = fdata[2][j];
    wscale[j] = fdata[3][j];
    extent[j] = fdata[4][j];
    radec_err[j] = fdata[5][j];
    ra[j]  = ddata[0][j];
    dec[j] = ddata[1][j];

  }//for

  for(unsigned int k = 0; k < i ; k++)
    delete col[k];
  dataSetServer->close(srclistSet);

  return 1;
}// readsrclist()
